
#pragma warning(disable:4996)

#include <windows.h>

#include <math.h>
#include <stdio.h>
#include <vector>
#include <list>

#include "lame.h"

#include "Socket.h"

#include "ShoutPlayer.h"

#pragma comment(lib, "wsock32")
#pragma comment(lib, "winmm")

static const int SAMPLING_RATE = 44100;
static const int BIT_RATE = 128;

class MP3Encoder {
private:
	lame_global_flags *_gf;
public:

	MP3Encoder()
	{
		_gf = 0;
	}

	~MP3Encoder()
	{
		close();
	}

	bool open()
	{
		_gf = lame_init();
		if (!_gf) {
			return false;
		}

		id3tag_init(_gf);
		lame_set_brate(_gf, BIT_RATE);
		lame_set_VBR_min_bitrate_kbps(_gf, lame_get_brate(_gf));
		lame_set_findReplayGain(_gf, 1);
		lame_set_num_channels(_gf, 2);
		lame_set_write_id3tag_automatic(_gf, 0);

		if (lame_init_params(_gf) < 0) {
			lame_close(_gf);
			return false;
		}

		return true;
	}

	struct encode_info_t {
		int in_samples;
		int *in_left;
		int *in_right;
		int out_length;
		unsigned char *out_mp3;
		encode_info_t()
			: in_samples(0)
			, in_left(0)
			, in_right(0)
			, out_length(0)
			, out_mp3(0)
		{
		}
	};

	bool encode(encode_info_t *buffer)
	{
		buffer->out_length = lame_encode_buffer_int(_gf, buffer->in_left, buffer->in_right, buffer->in_samples, buffer->out_mp3, LAME_MAXMP3BUFFER);
		return buffer->out_length >= 0;
	}

	bool flush(encode_info_t *buffer)
	{
		buffer->out_length = lame_encode_flush(_gf, buffer->out_mp3, LAME_MAXMP3BUFFER);
		return buffer->out_length >= 0;
	}

	void close()
	{
		if (_gf) {
			lame_close(_gf);
			_gf = 0;
		}
	}
};


class ShoutCastSource {
private:
	Socket _sock;
	std::string host;
	int port;
	std::string password;
public:
	ShoutCastSource()
	{
		port = 0;
	}
	void set_server_info(std::string const &h, int p, std::string const &pwd)
	{
		host = h;
		port = p;
		password = pwd;
	}
	bool send_song_name()
	{
		Socket sock;

		if (!sock.socket(PF_INET, SOCK_STREAM, 0)) {
			fprintf(stderr, "socket() failed\n");
			return false;
		}

		struct sockaddr_in serveraddr;
		{
			struct hostent *servhost;
			servhost = gethostbyname(host.c_str());
			if (servhost) {
				memcpy(&serveraddr.sin_addr, servhost->h_addr, servhost->h_length);
				serveraddr.sin_family = AF_INET;
				serveraddr.sin_port = htons((unsigned short)port);
				memset(serveraddr.sin_zero, 0, sizeof(serveraddr.sin_zero));
			}
		}

		if (!sock.connect((struct sockaddr *)&serveraddr, sizeof(serveraddr))) {
			fprintf(stderr,"connect() failed\n");
			sock.close();
			return false;
		}

		int n;
		char tmp[1000];

		sprintf(tmp, 
			"GET /admin.cgi?pass=%s&mode=updinfo&song=%s&url=%s HTTP/1.0\r\n"
			"User-Agent: ShoutPlayer(Mozilla Compatible)\r\n"
			"\r\n"
			, password.c_str()
			, "Unnamed%20Song"
			, ""
			);

		n = sock.send(tmp, (int)strlen(tmp), 0);
		if (n != strlen(tmp)) {
			return false;
		}

		sock.close();

		return true;
	}
	bool open()
	{
		if (port == 0) {
			return false;
		}
	
		int err;

		if (!_sock.socket(PF_INET, SOCK_STREAM, 0)) {
			fprintf(stderr, "socket() failed\n");
			return false;
		}

		struct sockaddr_in serveraddr;
		{
			struct hostent *servhost;
			servhost = gethostbyname(host.c_str());
			if (servhost) {
				memcpy(&serveraddr.sin_addr, servhost->h_addr, servhost->h_length);
				serveraddr.sin_family = AF_INET;
				serveraddr.sin_port = htons((unsigned short)(port + 1));
				memset(serveraddr.sin_zero, 0, sizeof(serveraddr.sin_zero));
			}
		}

		err = _sock.connect((struct sockaddr *)&serveraddr, sizeof(serveraddr));
		if (err == SOCKET_ERROR) {
			fprintf(stderr,"connect() failed\n");
			_sock.close();
			return false;
		}

		int n;
		char tmp[1000];

		sprintf(tmp, "%s\n", password.c_str());

		n = _sock.send(tmp, (int)strlen(tmp), 0);
		if (n != strlen(tmp)) {
			return false;
		}

		n = _sock.recv(tmp, sizeof(tmp), 0);
		if (n < 4) {
			return false;
		}
		if (tmp[0] == 'O' && tmp[1] == 'K' && tmp[2] == '2' && (tmp[3] == '\n' || tmp[3] == '\r')) {
			// ok
		} else {
			return false;
		}

		sprintf(tmp,
			"icy-name:Unnamed Server\n"
			"icy-genre:\n"
			"icy-url:\n"
			"icy-aim:\n"
			"icy-irc:\n"
			"icy-br:%u\n"
			"content-type:audio/mpeg\n"
			"\n"
			, BIT_RATE
			);

		n = _sock.send(tmp, (int)strlen(tmp), 0);
		if (n != strlen(tmp)) {
			return false;
		}

		send_song_name();

		return true;
	}

	void close()
	{
		_sock.close();
	}

	bool sendtcp(void const *ptr, size_t len)
	{
		if (!_sock.isvalid()) {
			return false;
		}
		return _sock.sendtcp(ptr, len);
	}

	bool connected() const
	{
		return _sock.isvalid();
	}
};



struct ShoutPlayer::Privates {
	enum {
		Closed,
		Stopped,
		Started,
		Stopping,
	} state;

	MP3Encoder encoder;
	ShoutCastSource source;

	HWAVEIN hWaveIn;
	WAVEFORMATEX wfe;
	std::vector<int> in_stream_left;
	std::vector<int> in_stream_right;
	struct Buffer {
		WAVEHDR whdr;
		std::vector<short> waveform;
	};
	std::list<Buffer> in_buffer_list;

	Privates()
	{
		state = Closed;
	}
};

ShoutPlayer::ShoutPlayer()
{
	privates = new Privates();
}

ShoutPlayer::~ShoutPlayer()
{
	delete privates;
}

int ShoutPlayer::get_buffer_length() const
{
	//return wfe.nAvgBytesPerSec / 10;
	return (1152 * 4) * 10;
}

bool ShoutPlayer::Open(int index)
{
	Close();

	privates->encoder.open();
	privates->source.open();

	privates->wfe.wFormatTag = WAVE_FORMAT_PCM;
	privates->wfe.nChannels = 2;
	privates->wfe.wBitsPerSample = 16;
	privates->wfe.nBlockAlign = privates->wfe.nChannels * privates->wfe.wBitsPerSample / 8;
	privates->wfe.nSamplesPerSec = SAMPLING_RATE;
	privates->wfe.nAvgBytesPerSec = privates->wfe.nSamplesPerSec * privates->wfe.nBlockAlign;

	index = WAVE_MAPPER;

	MMRESULT r = waveInOpen(&privates->hWaveIn, index, &privates->wfe, (DWORD_PTR)waveProc, (DWORD_PTR)this, CALLBACK_FUNCTION);
	if (r == MMSYSERR_NOERROR) {
		privates->state = Privates::Stopped;
		return true;
	}

	return false;
}

void ShoutPlayer::clear_buffer()
{
	privates->in_stream_left.clear();
	privates->in_stream_right.clear();
	privates->in_buffer_list.clear();
}

bool ShoutPlayer::Start()
{
	if (privates->state == Privates::Stopped) {
		clear_buffer();
		add_next_in_buffer();
		add_next_in_buffer();
		MMRESULT r = waveInStart(privates->hWaveIn);
		if (r == MMSYSERR_NOERROR) {
			privates->state = Privates::Started;
			return true;
		}
	}
	return false;
}
void ShoutPlayer::Stop()
{
	if (privates->state == Privates::Started) {
		privates->state = Privates::Stopping;
		while (!privates->in_buffer_list.empty()) {
			Sleep(1);
		}
		waveInStop(privates->hWaveIn);
		privates->state = Privates::Stopped;
	}
}
void ShoutPlayer::Close()
{
	Stop();

	if (privates->state != Privates::Closed) {
		waveInClose(privates->hWaveIn);
	}
	privates->state = Privates::Closed;

	privates->hWaveIn = 0;

	privates->source.close();
	privates->encoder.close();

	clear_buffer();
}

void ShoutPlayer::SetServerInfo(std::string const &host, int port, std::string const &password)
{
	privates->source.set_server_info(host, port, password);
}

bool ShoutPlayer::ConnectServer()
{
	return privates->source.open();
}

bool ShoutPlayer::ServerConnected() const
{
	return privates->source.connected();
}

void ShoutPlayer::add_next_in_buffer()
{
	std::list<Privates::Buffer>::iterator it = privates->in_buffer_list.insert(privates->in_buffer_list.end(), Privates::Buffer());

	int len = get_buffer_length();

	it->waveform.resize(len);

	it->whdr.lpData = (char *)&it->waveform[0];
	it->whdr.dwBufferLength = len;
	it->whdr.dwBytesRecorded = 0;
	it->whdr.dwFlags = 0;
	it->whdr.dwLoops = 1;
	it->whdr.lpNext = NULL;
	it->whdr.dwUser = 0;
	it->whdr.reserved = 0;

	waveInPrepareHeader(privates->hWaveIn, &it->whdr, sizeof(WAVEHDR));
	waveInAddBuffer(privates->hWaveIn, &it->whdr, sizeof(WAVEHDR));
}

void ShoutPlayer::wimData(HWAVEOUT hwo, DWORD_PTR dwParam1, DWORD_PTR dwParam2)
{
	short *waveptr = (short *)privates->in_buffer_list.front().whdr.lpData;
	int samples = privates->in_buffer_list.front().whdr.dwBufferLength / 4;

	for (int i = 0; i < samples; i++) {
		privates->in_stream_left.push_back(waveptr[0] * 65536);
		privates->in_stream_right.push_back(waveptr[1] * 65536);
		waveptr += 2;
	}

	size_t offset = 0;
	while (offset + 1152 <= privates->in_stream_left.size()) {
		unsigned char mp3[LAME_MAXMP3BUFFER];
		MP3Encoder::encode_info_t info;
		info.in_samples = 1152;
		info.in_left = &privates->in_stream_left[offset];
		info.in_right = &privates->in_stream_right[offset];
		info.out_mp3 = mp3;

		if (privates->encoder.encode(&info)) {
			privates->source.sendtcp(mp3, info.out_length);
		}

		offset += info.in_samples;
	}

	if (offset < privates->in_stream_left.size()) {
		size_t n = privates->in_stream_left.size() - offset;
		memmove(&privates->in_stream_left[0], &privates->in_stream_left[0] + privates->in_stream_left.size() - n, n * sizeof(int));
		memmove(&privates->in_stream_right[0], &privates->in_stream_right[0] + privates->in_stream_right.size() - n, n * sizeof(int));
		privates->in_stream_left.resize(n);
		privates->in_stream_right.resize(n);
	} else {
		privates->in_stream_left.clear();
		privates->in_stream_right.clear();
	}

	privates->in_buffer_list.pop_front();

	if (privates->state == Privates::Started) {
		add_next_in_buffer();
	} else {
		_asm nop
	}
}

void CALLBACK ShoutPlayer::waveProc(HWAVEOUT hwo, UINT uMsg, DWORD_PTR dwInstance, DWORD_PTR dwParam1, DWORD_PTR dwParam2)
{
	ShoutPlayer *me = (ShoutPlayer *)dwInstance;
	switch (uMsg) {
	//case WIM_OPEN:  me->wimOpen(hwo, dwParam1, dwParam2);  break;
	case WIM_DATA:  me->wimData(hwo, dwParam1, dwParam2);  break;
	//case WIM_CLOSE: me->wimClose(hwo, dwParam1, dwParam2); break;
	}
}

#if 0
int main2(int argc, char **argv)
{
	ShoutPlayer player;

	player.Open();
	player.Start();

	getchar();

	player.Stop();
	player.Close();

	return 0;
}
#endif
