
#pragma warning(disable:4996)

#include <windows.h>
#include "socket.h"

#ifdef WIN32
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#else
#include <fcntl.h>
#include <sys/stat.h>
#define O_BINARY 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>

#include <map>
#include <string>
#include <vector>
#include <sstream>

#include "CombinePath.h"
#include "sswrite.h"


#include "httpserver.h"

namespace httpserver {

	struct SocketBuffer {
		Socket sock;
		std::vector<unsigned char> buffer;
		int offset;
		int length;
		SocketBuffer()
		{
			offset = 0;
			length = 0;
		}
		std::string readline()
		{
			if (buffer.empty()) {
				buffer.resize(1024);
			}
			while (true) {
				if (offset < length) {
					unsigned char *left = &buffer[offset];
					unsigned char *right = (unsigned char *)memchr(left, '\n', length - offset);
					if (right) {
						offset = right + 1 - &buffer[0];
						while (left < right && isspace(right[-1])) {
							right--;
						}
						return std::string(left, right);
					}
					int n = length - offset;
					memmove(&buffer[0], left, n);
					length = n;
					offset = 0;
				} else {
					length = 0;
					offset = 0;
				}
				int l = sock.recv((char *)&buffer[0] + length, buffer.size() - length, 0);
				if (l < 1) {
					return std::string();
				}
				length += l;
			}
		}
	};

	static void split(std::string const &line, std::vector<std::string> *vec)
	{
		char const *ptr = line.c_str();
		char const *end = ptr + line.size();
		while (isspace(*ptr)) {
			ptr++;
		}
		char const *left = ptr;
		while (true) {
			if (left) {
				if (!*ptr || isspace(*ptr)) {
					std::string line(left, ptr);
					vec->push_back(line);
					if (!*ptr) {
						return;
					}
					left = 0;
				}
			} else {
				if (!isspace(*ptr)) {
					left = ptr;
				}
			}
			ptr++;
		}
	}

	void http_response_t::set_content_type(std::string const &ct)
	{
		std::stringstream ss;
		sswrite(&ss, "Content-Type: ");
		sswrite(&ss, ct);
		header.push_back(ss.str().c_str());
	}

	http_status_t const *http_response_t::set_location(std::string const &url)
	{
		std::stringstream ss;
		sswrite(&ss, "Location: ");
		sswrite(&ss, url);
		std::string str = ss.str();
		header.push_back(str);
		return http_303_see_other;
	}

	// HTTP_Server

	struct HTTP_Thread : public Thread {
		HTTP_Server *server;
		void setup(HTTP_Server *s)
		{
			server = s;
		}
		virtual void run()
		{
			try {
				while (true) {
					struct sockaddr_in peer_sin;

					int len = sizeof(peer_sin);

					Socket connected_socket = server->listener_socket.accept((struct sockaddr *)&peer_sin, &len);
					if (!connected_socket.isvalid()) {
						throw "accept failed";
					}

					SocketBuffer sockbuff;
					sockbuff.sock = connected_socket;

					http_request_t request;

					while (1) {
						std::string line = sockbuff.readline();
						if (line.empty()) {
							if (request.header.size() > 0) {
								http_response_t response;
								http_status_t const *status = server->http_process_request(&request, &response);
								if (status->code / 100 >= 4) {
									if (response.buffer.empty()) {
										char tmp[10];
										sprintf(tmp, "%u ", status->code);
										unsigned char const *p;
										p = (unsigned char const *)tmp;
										response.buffer.insert(response.buffer.end(), p, p + strlen(tmp));
										p = (unsigned char const *)status->text;
										response.buffer.insert(response.buffer.end(), p, p + strlen(status->text));
										response.set_content_type("text/plain");
									}
								}
								server->http_send_response_header(sockbuff.sock, status, &response);
								if (!response.buffer.empty()) {
									sockbuff.sock.send((char const *)&response.buffer[0], response.buffer.size(), 0);
								}
							}
							break;
						}
						request.header.push_back(line);
					}

					connected_socket.close();
				}
			} catch (char const *) {
			}
		}
	};

	HTTP_Server::HTTP_Server()
	{
		tcp_port = 8080;
//		document_root = "/var/www";
	}



	static bool validate_uri(std::string const &uri)
	{
		char const *p = uri.c_str();
		if (strchr(p, '\\')) {
			return false;
		}
		if (strstr(p, "..")) {
			return false;
		}
		if (strstr(p, "//")) {
			return false;
		}
		if (strstr(p, "/.")) {
			return false;
		}
		return true;
	}

	http_status_t const *HTTP_Server::default_handler(std::string const &uri, std::string const &suffix, http_response_t *response)
	{
		std::string mimetype;
		if (!suffix.empty()) {
			mimetype = content_type_map.find(suffix);
		}
		if (mimetype.empty()) {
			response->set_content_type("application/octet-stream");
		} else {
			response->set_content_type(mimetype.c_str());
		}
		std::string path = CombinePath(document_root, uri.c_str() + 1);
		response->buffer.clear();
		struct stat st;
		if (stat(path.c_str(), &st) == 0) {
			int fd = ::open(path.c_str(), O_RDONLY | O_BINARY);
			if (fd != -1) {
				response->buffer.resize(st.st_size);
				if (read(fd, &response->buffer[0], st.st_size) != st.st_size) {
					::close(fd);
					return http_500_internal_server_error;
				}
				::close(fd);
				return http_200_ok;
			}
		}
		return http_404_not_found;
	}

	http_status_t const *HTTP_Server::dispatch_get_request(std::string const &uri, http_request_t *request, http_response_t *response)
	{
		for (std::vector<http_get_handler_item_t>::const_iterator it = http_get_handler_table.begin(); it != http_get_handler_table.end(); it++) {
			http_status_t const *s = it->handler(it->cookie, uri, request, response);
			if (s) {
				return s;
			}
		}
		return 0;
	}

	http_status_t const *HTTP_Server::http_process_request(http_request_t *request, http_response_t *response)
	{
		std::vector<std::string> first;
		split(request->header[0], &first);
		if (first.size() < 3) {
			return http_400_bad_request;
		}
		if (first[0] == "GET") {
			std::string uri = first[1];
			if (!validate_uri(uri)) {
				return http_400_bad_request;
			}
			if (uri.size() > 0) {
				if (uri[0] == '/') {
					{
						http_status_t const *s = dispatch_get_request(uri, request, response);
						if (s) {
							return s;
						}
					}
					if (uri[uri.size() - 1] == '/') {
						uri += "index.html";
						std::string path = CombinePath(document_root, uri.c_str() + 1);
						struct stat st;
						if (stat(path.c_str(), &st) == 0) {
							return response->set_location(uri);
						}
						return http_403_forbidden;
					} else {
						std::string suffix;
						{
							char const *p = strrchr(uri.c_str(), '.');
							if (p) {
								suffix = p + 1;
							}
						}
						{
							SuffixHandler *h = suffix_handler_map.find(suffix);
							if (h) {
								http_status_t const *s = h->process(uri, suffix, response);
								if (s) {
									return s;
								}
								return http_400_bad_request;
							}
						}
						return default_handler(uri, suffix, response);
					}
				}
			}
			return http_400_bad_request;
		}
		return http_405_method_not_allowed;
	}

	void HTTP_Server::http_send_response_header(Socket sock, http_status_t const *status, http_response_t *response)
	{
		std::vector<unsigned char> out;
		out.reserve(1024);
		{
			char tmp[100];
			sprintf(tmp, "HTTP/1.1 %03u %s\r\n", status->code, status->text);
			out.insert(out.end(), tmp, tmp + strlen(tmp));
		}
		for (std::vector<std::string>::const_iterator it = response->header.begin(); it != response->header.end(); it++) {
			unsigned char const *p = (unsigned char const *)it->c_str();
			out.insert(out.end(), p, p + it->size());
			out.push_back('\r');
			out.push_back('\n');
		}
		out.push_back('\r');
		out.push_back('\n');
		sock.send((char const *)&out[0], out.size(), 0);
	}

	void HTTP_Server::run()
	{
		try {
			std::vector<HTTP_Thread> threads;

			int ret;

			struct sockaddr_in sin;

			listener_socket.socket(AF_INET, SOCK_STREAM, 0);
			if (!listener_socket.isvalid()) {
				throw "socket() failed";
			}

			{
				int val = 1;
				if (listener_socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, (char const *)&val, sizeof(val)) == -1) {
					throw "setsockopt() failed";
				}
			}

			sin.sin_family = AF_INET;
			sin.sin_port = htons(tcp_port);
			sin.sin_addr.s_addr = htonl(INADDR_ANY);

			if (listener_socket.bind((struct sockaddr *)&sin, sizeof(sin)) < 0) {
				throw "bind() failed";
			}

			ret = listener_socket.listen(SOMAXCONN);
			if (ret == -1) {
				throw "listen() failed";
			}

			int threadcount = 10;
			threads.resize(threadcount);
			for (int i = 0; i < threadcount; i++) {
				threads[i].setup(this);
			}
			for (int i = 0; i < threadcount; i++) {
				threads[i].start();
			}

			while (!interrupted()) {
				listener_event.wait(1000);
			}

			listener_socket.close();

			for (int i = 0; i < threadcount; i++) {
				threads[i].join();
			}

		} catch (char const *msg) {
			fprintf(stderr, "%s\n", msg);
		}
	}

	void HTTP_Server::stop()
	{
		Thread::stop();
		listener_event.signal();
	}
}

//

